@@ -1,12 +1,10 @@
-use super::shared::{SpinWait, Waiter};
-#[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
-use lock_api::RawRwLock as _RawRwLock;
+#![allow(unused_imports)]
+use super::shared::{fence_acquire, invalid_mut, AtomicPtrRmw, SpinWait, StrictProvenance, Waiter};
 use std::{
     fmt,
-    marker::PhantomData,
     pin::Pin,
-    ptr::NonNull,
-    sync::atomic::{fence, AtomicUsize, Ordering},
+    ptr::{self, NonNull},
+    sync::atomic::{AtomicPtr, Ordering},
 };
 
 const UNLOCKED: usize = 0;
@@ -53,8 +51,7 @@ pub struct RawRwLock {
     ///        |         |        |              |           | the head Waiter node of the waiting thread queue. There is
     ///        |         |        |              |           | also a thread which is updating the waiting-thread queue.
     /// -------+---------+--------+--------------+-----------+-------------------------------------------------------------
-    pub(super) state: AtomicUsize,
-    _p: PhantomData<*const Waiter>,
+    pub(super) state: AtomicPtr<Waiter>,
 }
 
 impl fmt::Debug for RawRwLock {
@@ -70,20 +67,19 @@ unsafe impl lock_api::RawRwLock for RawRwLock {
     type GuardMarker = crate::GuardMarker;
 
     const INIT: Self = Self {
-        state: AtomicUsize::new(UNLOCKED),
-        _p: PhantomData,
+        state: AtomicPtr::new(invalid_mut(UNLOCKED)),
     };
 
     #[inline]
     fn is_locked(&self) -> bool {
         let state = self.state.load(Ordering::Relaxed);
-        state & LOCKED != 0
+        state.address() & LOCKED != 0
     }
 
     #[inline]
     fn is_locked_exclusive(&self) -> bool {
         let state = self.state.load(Ordering::Relaxed);
-        state & (LOCKED | READING) == LOCKED
+        state.address() & (LOCKED | READING) == LOCKED
     }
 
     #[inline]
@@ -125,10 +121,11 @@ unsafe impl lock_api::RawRwLock for RawRwLock {
 
 //  --- X86 Specializations
 
-#[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
+#[cfg(all(any(target_arch = "x86", target_arch = "x86_64"), not(miri)))]
 impl RawRwLock {
     #[inline(always)]
-    fn try_lock_exclusive_assuming(&self, _state: usize) -> bool {
+    fn try_lock_exclusive_assuming(&self, _state: *mut Waiter) -> bool {
+        use lock_api::RawRwLock as _;
         self.try_lock_exclusive()
     }
 
@@ -155,7 +152,11 @@ impl RawRwLock {
                 out(reg_byte) old_locked_bit,
                 options(nostack),
             );
-            old_locked_bit == 0
+            let acquired = old_locked_bit == 0;
+            if acquired {
+                fence_acquire(&self.state);
+            }
+            acquired
         }
     }
 
@@ -164,11 +165,12 @@ impl RawRwLock {
         // On x86, we unlock the exclusive lock first, then try and wake later.
         // This is faster than using a `lock cmpxchg` loop as it doesn't have
         // to fail and retry from other threads updating QUEUE_LOCKED bit or queueing themselves.
-        let state = self.state.fetch_sub(LOCKED, Ordering::Release);
-        debug_assert_eq!(state & (LOCKED | READING), LOCKED);
+        let locked = ptr::null_mut::<Waiter>().with_address(LOCKED);
+        let state = self.state.fetch_sub(locked, Ordering::Release);
+        debug_assert_eq!(state.address() & (LOCKED | READING), LOCKED);
 
         // Only try to unpark if there's no QUEUE_LOCKED owner yet and if there's threads queued.
-        if state & (QUEUED | QUEUE_LOCKED) == QUEUED {
+        if state.address() & (QUEUED | QUEUE_LOCKED) == QUEUED {
             self.try_unpark();
         }
     }
@@ -178,11 +180,12 @@ impl RawRwLock {
         // On x86, we unlock the shared lock first, then try and wake later.
         // This is faster than using a `lock cmpxchg` loop as it doesn't have
         // to fail and retry from other threads updating QUEUE_LOCKED bit or queueing themselves.
-        let state = self.state.fetch_sub(LOCKED | READING, Ordering::Release);
-        debug_assert_eq!(state & (LOCKED | READING), LOCKED | READING);
+        let read_locked = ptr::null_mut::<Waiter>().with_address(LOCKED | READING);
+        let state = self.state.fetch_sub(read_locked, Ordering::Release);
+        debug_assert_eq!(state.address() & (LOCKED | READING), LOCKED | READING);
 
         // Only try to unpark if there's no QUEUE_LOCKED owner yet and if there's threads queued.
-        if state & (QUEUED | QUEUE_LOCKED) == QUEUED {
+        if state.address() & (QUEUED | QUEUE_LOCKED) == QUEUED {
             self.try_unpark();
         }
     }
@@ -195,8 +198,8 @@ impl RawRwLock {
         // - theres no lock holder, as they can be the ones to do the wake up
         // - there are still threads queued to actually wake up
         // - the QUEUE_LOCKED bit isnt held as someone is already doing wake up
-        while state & (LOCKED | QUEUED | QUEUE_LOCKED) == QUEUED {
-            let new_state = state | QUEUE_LOCKED;
+        while state.address() & (LOCKED | QUEUED | QUEUE_LOCKED) == QUEUED {
+            let new_state = state.map_address(|addr| addr | QUEUE_LOCKED);
             match self.state.compare_exchange_weak(
                 state,
                 new_state,
@@ -210,14 +213,14 @@ impl RawRwLock {
     }
 }
 
-#[cfg(not(any(target_arch = "x86", target_arch = "x86_64")))]
+#[cfg(any(miri, not(any(target_arch = "x86", target_arch = "x86_64"))))]
 impl RawRwLock {
     #[inline(always)]
-    fn try_lock_exclusive_assuming(&self, mut state: usize) -> bool {
-        while state & LOCKED == 0 {
+    fn try_lock_exclusive_assuming(&self, mut state: *mut Waiter) -> bool {
+        while state.address() & LOCKED == 0 {
             match self.state.compare_exchange_weak(
                 state,
-                state | LOCKED,
+                state.map_address(|addr| addr | LOCKED),
                 Ordering::Acquire,
                 Ordering::Relaxed,
             ) {
@@ -232,7 +235,12 @@ impl RawRwLock {
     #[inline(always)]
     fn try_lock_exclusive_fast(&self) -> bool {
         self.state
-            .compare_exchange(UNLOCKED, LOCKED, Ordering::Acquire, Ordering::Relaxed)
+            .compare_exchange(
+                invalid_mut(UNLOCKED),
+                invalid_mut(LOCKED),
+                Ordering::Acquire,
+                Ordering::Relaxed,
+            )
             .is_ok()
     }
 
@@ -240,7 +248,12 @@ impl RawRwLock {
     unsafe fn unlock_exclusive_fast(&self) {
         if self
             .state
-            .compare_exchange(LOCKED, UNLOCKED, Ordering::Release, Ordering::Relaxed)
+            .compare_exchange(
+                invalid_mut(LOCKED),
+                invalid_mut(UNLOCKED),
+                Ordering::Release,
+                Ordering::Relaxed,
+            )
             .is_err()
         {
             self.unlock_and_unpark();
@@ -256,17 +269,15 @@ impl RawRwLock {
     unsafe fn unlock_and_unpark(&self) {
         let mut state = self.state.load(Ordering::Relaxed);
         loop {
-            assert_ne!(state & LOCKED, 0);
-            assert_ne!(state & QUEUED, 0);
+            assert_ne!(state.address() & LOCKED, 0);
+            assert_ne!(state.address() & QUEUED, 0);
 
             // Unlocks the rwlock and tries to grab the QUEUE_LOCKED bit for wake up.
-            let mut new_state = state & !LOCKED;
-            new_state |= QUEUE_LOCKED;
-
-            // Also unset the READING bit if the caller was the last shared owner
-            if state & READING != 0 {
-                new_state &= !READING;
-            }
+            let new_state = state.map_address(|mut addr| {
+                addr &= !(LOCKED | READING);
+                addr |= QUEUE_LOCKED;
+                addr
+            });
 
             if let Err(e) = self.state.compare_exchange_weak(
                 state,
@@ -278,7 +289,7 @@ impl RawRwLock {
                 continue;
             }
 
-            if state & QUEUE_LOCKED == 0 {
+            if state.address() & QUEUE_LOCKED == 0 {
                 self.unpark(new_state);
             }
 
@@ -291,10 +302,13 @@ impl RawRwLock {
 
 impl RawRwLock {
     #[inline(always)]
-    fn try_lock_shared_assuming(&self, state: usize) -> Option<Result<usize, usize>> {
+    fn try_lock_shared_assuming(
+        &self,
+        state: *mut Waiter,
+    ) -> Option<Result<*mut Waiter, *mut Waiter>> {
         // Returns None if the lock is held by a writer
-        if state != UNLOCKED {
-            if state & (LOCKED | READING | QUEUED) != (LOCKED | READING) {
+        if state.address() != UNLOCKED {
+            if state.address() & (LOCKED | READING | QUEUED) != (LOCKED | READING) {
                 return None;
             }
         }
@@ -304,10 +318,10 @@ impl RawRwLock {
         // Overflow is very unlikely though as it requires `usize::MAX >> 4` active readers at once.
         // On a system where `usize` is 64 bits, that's over a quintillion (1 million ^ 5) readers.
         // On a system where `usize` is 32 bits, that's still over 260 million readers.
-        if let Some(with_reader) = state.checked_add(1 << READER_SHIFT) {
+        if let Some(with_reader) = state.address().checked_add(1 << READER_SHIFT) {
             return Some(self.state.compare_exchange_weak(
                 state,
-                with_reader | LOCKED | READING,
+                state.with_address(with_reader | LOCKED | READING),
                 Ordering::Acquire,
                 Ordering::Relaxed,
             ));
@@ -339,14 +353,14 @@ impl RawRwLock {
     unsafe fn unlock_shared_fast(&self) -> bool {
         // Just go to the slow path if we're not the only reader
         let state = self.state.load(Ordering::Relaxed);
-        if state != SINGLE_READER {
+        if state.address() != SINGLE_READER {
             return false;
         }
 
         self.state
             .compare_exchange(
-                SINGLE_READER,
-                UNLOCKED,
+                state.with_address(SINGLE_READER),
+                state.with_address(UNLOCKED),
                 Ordering::Release,
                 Ordering::Relaxed,
             )
@@ -359,14 +373,14 @@ impl RawRwLock {
         // This only works because the Remaining bits still point to the reader count.
         // When threads start waiting, they override these bits with the queue pointer.
         let mut state = self.state.load(Ordering::Relaxed);
-        while state & QUEUED == 0 {
-            assert_ne!(state & LOCKED, 0);
-            assert_ne!(state & READING, 0);
-            assert_ne!(state >> READER_SHIFT, 0);
+        while state.address() & QUEUED == 0 {
+            assert_ne!(state.address() & LOCKED, 0);
+            assert_ne!(state.address() & READING, 0);
+            assert_ne!(state.address() >> READER_SHIFT, 0);
 
-            let mut new_state = state - (1 << READER_SHIFT);
-            if state == SINGLE_READER {
-                new_state = UNLOCKED;
+            let mut new_state = state.map_address(|addr| addr - (1 << READER_SHIFT));
+            if state.address() == SINGLE_READER {
+                new_state = state.with_address(UNLOCKED);
             }
 
             match self.state.compare_exchange_weak(
@@ -382,15 +396,15 @@ impl RawRwLock {
 
         // The'ers threads waiting on the RwLock.
         // The reader count has moved to the tail of the queue.
-        assert_ne!(state & LOCKED, 0);
-        assert_ne!(state & QUEUED, 0);
-        assert_ne!(state & READING, 0);
+        assert_ne!(state.address() & LOCKED, 0);
+        assert_ne!(state.address() & QUEUED, 0);
+        assert_ne!(state.address() & READING, 0);
 
         // Find the tail of the wait queue while also caching it at the current head.
         // As long as the Waiter writes are atomic, this can be soundly racing with
         // other callers to get_and_link_queue() like link_queue_or_unpark() or other readers.
         // Acquire barrier to ensure Waiter queue writes to head happen before we start scanning.
-        fence(Ordering::Acquire);
+        fence_acquire(&self.state);
         let (_head, tail) = Waiter::get_and_link_queue(state, |_| {});
 
         // Decrement the reader count which was moved to the tail.
@@ -402,7 +416,7 @@ impl RawRwLock {
         // Acquire barrier synchronizes with the Release to counter above to ensure
         // that the unsetting of the LOCKED bit happens after all the readers reads/loads occur.
         if readers == 1 {
-            fence(Ordering::Acquire);
+            fence_acquire(&self.state);
             self.unlock_shared_and_unpark();
         }
     }
@@ -410,8 +424,8 @@ impl RawRwLock {
     #[cold]
     fn lock_exclusive_slow(&self) {
         let is_writer = true;
-        let try_lock = |state: usize| -> Option<bool> {
-            match state & LOCKED {
+        let try_lock = |state: *mut Waiter| -> Option<bool> {
+            match state.address() & LOCKED {
                 0 => Some(self.try_lock_exclusive_assuming(state)),
                 _ => None,
             }
@@ -423,7 +437,7 @@ impl RawRwLock {
     #[cold]
     fn lock_shared_slow(&self) {
         let is_writer = false;
-        let try_lock = |state: usize| -> Option<bool> {
+        let try_lock = |state: *mut Waiter| -> Option<bool> {
             let result = self.try_lock_shared_assuming(state)?;
             Some(result.is_ok())
         };
@@ -431,7 +445,7 @@ impl RawRwLock {
         self.lock_common(is_writer, try_lock)
     }
 
-    fn lock_common(&self, is_writer: bool, mut try_lock: impl FnMut(usize) -> Option<bool>) {
+    fn lock_common(&self, is_writer: bool, mut try_lock: impl FnMut(*mut Waiter) -> Option<bool>) {
         Waiter::with(|waiter| {
             waiter.waiting_on.set(Some(NonNull::from(self).cast()));
             waiter.flags.set(is_writer as usize);
@@ -455,7 +469,7 @@ impl RawRwLock {
                     // We can't acquire the RwLock at the moment.
                     // Try to spin for a little in hopes the RwLock is released quickly.
                     // Also don't spin if threads are waiting as we should start waiting too.
-                    if (state & QUEUED == 0) && spin.try_yield_now() {
+                    if (state.address() & QUEUED == 0) && spin.try_yield_now() {
                         state = self.state.load(Ordering::Relaxed);
                         continue;
                     }
@@ -480,7 +494,7 @@ impl RawRwLock {
         let mut state = self.state.load(Ordering::Relaxed);
         loop {
             // Don't requeue if the waiter (which is a writer) could acquire the lock.
-            if state & LOCKED == 0 {
+            if state.address() & LOCKED == 0 {
                 return false;
             }
 
@@ -491,15 +505,18 @@ impl RawRwLock {
         }
     }
 
-    unsafe fn try_queue(&self, state: &mut usize, waiter: Pin<&Waiter>) -> bool {
+    unsafe fn try_queue(&self, state: &mut *mut Waiter, waiter: Pin<&Waiter>) -> bool {
         // Prepare to push our waiter to the head of the wait queue.
-        let waiter_ptr = &*waiter as *const Waiter as usize;
-        let mut new_state = (*state & !Waiter::MASK) | waiter_ptr | QUEUED;
+        let waiter_ptr = NonNull::from(&*waiter).as_ptr();
+        let mut new_state = waiter_ptr.map_address(|addr| {
+            let state_bits = (*state).address() & !Waiter::MASK;
+            addr | state_bits | QUEUED
+        });
 
-        if *state & QUEUED == 0 {
+        if (*state).address() & QUEUED == 0 {
             // The first queued waiter will be the tail and it now needs to
             // track the readers since its overriding the remaining state bits.
-            let readers = *state >> READER_SHIFT;
+            let readers = (*state).address() >> READER_SHIFT;
             waiter.counter.store(readers, Ordering::Relaxed);
 
             // The first queued waiter also sets its `tail` field to itself.
@@ -510,8 +527,8 @@ impl RawRwLock {
         } else {
             // The thread which holds the QUEUE_LOCKED bit, or active read-lock holders, will update the queue.
             // Since there's multiple waiting threads now, try to grab the QUEUE_LOCKED bit in order to update the queue.
-            let head = NonNull::new((*state & Waiter::MASK) as *mut Waiter);
-            new_state |= QUEUE_LOCKED;
+            let head = NonNull::new((*state).map_address(|addr| addr & Waiter::MASK));
+            new_state = new_state.map_address(|addr| addr | QUEUE_LOCKED);
 
             // Other waiters will link themselves onto the waiter queue in a stack-like fashion;
             // Leaving the `tail` field unset for Waiter::get_and_link_queue() to traverse and cache the found tail.
@@ -532,7 +549,7 @@ impl RawRwLock {
             return false;
         }
 
-        if *state & (QUEUED | QUEUE_LOCKED) == QUEUED {
+        if (*state).address() & (QUEUED | QUEUE_LOCKED) == QUEUED {
             self.link_queue_or_unpark(new_state);
         }
 
@@ -540,29 +557,29 @@ impl RawRwLock {
     }
 
     #[cold]
-    unsafe fn link_queue_or_unpark(&self, mut state: usize) {
+    unsafe fn link_queue_or_unpark(&self, mut state: *mut Waiter) {
         loop {
-            assert_ne!(state & QUEUED, 0);
-            assert_ne!(state & QUEUE_LOCKED, 0);
+            assert_ne!(state.address() & QUEUED, 0);
+            assert_ne!(state.address() & QUEUE_LOCKED, 0);
 
             // If the lock holders released the lock,
             // we are now in charge of waking up threads since we hold the QUEUE_LOCKED bit.
             // This is due to the lock-releasing thread skipping thread-wakeup
             // if the QUEUE_LOCKED bit is set as we can take over its job.
-            if state & LOCKED == 0 {
+            if state.address() & LOCKED == 0 {
                 return self.unpark(state);
             }
 
             // Fix the prev links in the waiter queue now that we hold the QUEUE_LOCKED bit.
             // Acquire barrier to ensure writes to waiters pushed to the queue happen before we start fixing it.
-            fence(Ordering::Acquire);
+            fence_acquire(&self.state);
             let _ = Waiter::get_and_link_queue(state, |_| {});
 
             // Finally, try to the release the QUEUE_LOCKED bit.
             // Release barrier to ensure the writes we did above happen before the next QUEUE_LOCKED bit holder.
             match self.state.compare_exchange_weak(
                 state,
-                state & !QUEUE_LOCKED,
+                state.map_address(|addr| addr & !QUEUE_LOCKED),
                 Ordering::Release,
                 Ordering::Relaxed,
             ) {
@@ -573,19 +590,19 @@ impl RawRwLock {
     }
 
     #[cold]
-    unsafe fn unpark(&self, mut state: usize) {
+    unsafe fn unpark(&self, mut state: *mut Waiter) {
         loop {
-            assert_ne!(state & QUEUED, 0);
-            assert_ne!(state & QUEUE_LOCKED, 0);
+            assert_ne!(state.address() & QUEUED, 0);
+            assert_ne!(state.address() & QUEUE_LOCKED, 0);
 
             // If the RwLock is locked by another thread while we're trying to wake one up,
             // then bail by releasing the QUEUE_LOCKED bit as the lock holder can do the wake up instead.
             // Release barrier to ensure the queue writes we've possibly done so far in Waiter::get_and_link_queue()
             // below happen before the next QUEUE_LOCKED bit holder.
-            if state & LOCKED != 0 {
+            if state.address() & LOCKED != 0 {
                 match self.state.compare_exchange_weak(
                     state,
-                    state & !QUEUE_LOCKED,
+                    state.map_address(|addr| addr & !QUEUE_LOCKED),
                     Ordering::Release,
                     Ordering::Relaxed,
                 ) {
@@ -598,7 +615,7 @@ impl RawRwLock {
             // Fix and get the ends of the wait queue in order to wake the tail up.
             // Acquire barrier ensures that writes to waiters pushed to the queue
             // happen before we start fixing/getting it.
-            fence(Ordering::Acquire);
+            fence_acquire(&self.state);
             let (head, tail) = Waiter::get_and_link_queue(state, |_| {});
 
             // If the tail (the waiter to wake up) is a writer,
@@ -611,7 +628,8 @@ impl RawRwLock {
                     // Unset the QUEUE_LOCKED bit now that we have dequeued the tail for waking.
                     // Release barrier ensures the head/tail updates happen before the next QUEUE_LOCKED bit owner.
                     head.as_ref().tail.set(Some(new_tail));
-                    self.state.fetch_and(!QUEUE_LOCKED, Ordering::Release);
+                    self.state
+                        .fetch_sub(state.with_address(QUEUE_LOCKED), Ordering::Release);
 
                     // unpark_waiters() follows the queue backwards from the tail to the head using the `prev` field.
                     // Since we queue to the head, we dequeue from the tail.
@@ -631,7 +649,7 @@ impl RawRwLock {
             // Release barrier ensures the head/tail access above happen before we release the QUEUE_LOCKED bit before wake up.
             match self.state.compare_exchange_weak(
                 state,
-                state & !(Waiter::MASK | QUEUED | QUEUE_LOCKED),
+                state.map_address(|addr| addr & !(Waiter::MASK | QUEUED | QUEUE_LOCKED)),
                 Ordering::Release,
                 Ordering::Relaxed,
             ) {
@@ -797,7 +815,7 @@ mod tests {
     #[test]
     fn frob() {
         const N: u32 = 10;
-        const M: u32 = 1000;
+        const M: u32 = if cfg!(miri) { 100 } else { 1000 };
 
         let r = Arc::new(RwLock::new(()));
 
